# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


requirements = ["capstone"]

if sys.version_info >= (3, 0):
    # py3
    requirements.append("lief>=0.11")
else:
    # py2 - newer LIEF is Python3 only
    requirements.append("lief==0.9.0")


setup(
    name='smda',
    # note to self: always change this in config as well.
    version='1.8.1',
    description='A recursive disassmbler optimized for CFG recovery from memory dumps. Based on capstone.',
    long_description_content_type="text/markdown",
    long_description=long_description,
    author='Daniel Plohmann',
    author_email='daniel.plohmann@mailbox.org',
    url='https://github.com/danielplohmann/smda',
    license="BSD 2-Clause",
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=requirements,
    data_files=[
        ('', ['LICENSE']),
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Security",
        "Topic :: Software Development :: Disassemblers",
    ],
)





class DisassemblyStatistics(object):

    num_functions = None
    num_recursive_functions = None
    num_leaf_functions = None
    num_basic_blocks = None
    num_instructions = None
    num_api_calls = None
    num_function_calls = None
    num_failed_functions = None
    num_failed_instructions = None

    def __init__(self, disassembly_result=None):
        if disassembly_result is not None:
            self.num_functions = len(disassembly_result.functions)
            self.num_recursive_functions = len(disassembly_result.recursive_functions)
            self.num_leaf_functions = len(disassembly_result.leaf_functions)
            self.num_basic_blocks = self._countBlocks(disassembly_result)
            self.num_instructions = self._countInstructions(disassembly_result)
            self.num_api_calls = self._countApiCalls(disassembly_result)
            self.num_function_calls = self._countFunctionCalls(disassembly_result)
            self.num_failed_functions = len(disassembly_result.failed_analysis_addr)
            self.num_failed_instructions = len(disassembly_result.errors)

    def _countBlocks(self, disassembly_result):
        num_blocks = 0
        for _, blocks in disassembly_result.functions.items():
            num_blocks += len(blocks)
        return num_blocks

    def _countApiCalls(self, disassembly_result):
        return len(disassembly_result.getAllApiRefs())

    def _countInstructions(self, disassembly_result):
        num_ins = 0
        for function_offset in sorted(disassembly_result.functions):
            for block in disassembly_result.functions[function_offset]:
                num_ins += len(block)
        return num_ins

    def _countFunctionCalls(self, disassembly_result):
        num_calls = 0
        for function_start in disassembly_result.functions:
            if function_start in disassembly_result.code_refs_to:
                num_calls += len(disassembly_result.code_refs_to[function_start])
        return num_calls

    @classmethod
    def fromDict(cls, statistics_dict):
        statistics = cls(None)
        statistics.num_functions = statistics_dict["num_functions"]
        statistics.num_recursive_functions = statistics_dict["num_recursive_functions"]
        statistics.num_leaf_functions = statistics_dict["num_leaf_functions"]
        statistics.num_basic_blocks = statistics_dict["num_basic_blocks"]
        statistics.num_instructions = statistics_dict["num_instructions"]
        statistics.num_api_calls = statistics_dict["num_api_calls"]
        statistics.num_function_calls = statistics_dict["num_function_calls"]
        statistics.num_failed_functions = statistics_dict["num_failed_functions"]
        statistics.num_failed_instructions = statistics_dict["num_failed_instructions"]
        return statistics

    def toDict(self):
        return {
            "num_functions": self.num_functions,
            "num_recursive_functions": self.num_recursive_functions,
            "num_leaf_functions": self.num_leaf_functions,
            "num_basic_blocks": self.num_basic_blocks,
            "num_instructions": self.num_instructions,
            "num_api_calls": self.num_api_calls,
            "num_function_calls": self.num_function_calls,
            "num_failed_functions": self.num_failed_functions,
            "num_failed_instructions": self.num_failed_instructions
        }

    def __add__(self, other):
        if not isinstance(other, DisassemblyStatistics):
            raise ValueError("Needs another DisassemblyStatistics to perform addition of values")
        self.num_functions += other.num_functions
        self.num_recursive_functions += other.num_recursive_functions
        self.num_leaf_functions += other.num_leaf_functions
        self.num_basic_blocks += other.num_basic_blocks
        self.num_instructions += other.num_instructions
        self.num_api_calls += other.num_api_calls
        self.num_function_calls += other.num_function_calls
        self.num_failed_functions += other.num_failed_functions
        self.num_failed_instructions += other.num_failed_instructions
        return self



from binascii import hexlify

from .definitions import COMMON_PROLOGUES

class FunctionCandidate(object):

    def __init__(self, binary_info, addr):
        self.bitness = binary_info.bitness
        self.addr = addr
        rel_start_addr = addr - binary_info.base_addr
        self.bytes = binary_info.binary[rel_start_addr:rel_start_addr + 5]
        self.lang_spec = None
        self.call_ref_sources = []
        self.finished = False
        self.is_symbol = False
        self.is_gap_candidate = False
        self.is_tailcall = False
        self.alignment = 0
        if addr % 4 == 0:
            self.alignment = 4
        elif addr % 16 == 0:
            self.alignment = 16
        self.analysis_aborted = False
        self.abortion_reason = ""
        self._score = None
        self._tfidf_score = None
        self._confidence = None
        self.function_start_score = None
        self.is_stub = False
        self.is_initial_candidate = False
        self.is_exception_handler = False

    def setTfIdf(self, tfidf_score):
        self._tfidf_score = tfidf_score

    def getTfIdf(self):
        return round(self._tfidf_score, 3)

    def getConfidence(self):
        if self._confidence is None:
            # based on evaluation over Andriesse, Bao, and Plohmann data sets
            weighted_confidence = 0.298 * (1 if self.hasCommonFunctionStart() else 0)
            if self._tfidf_score is not None:
                weighted_confidence += (
                    0.321 * (1 if self._tfidf_score < 0 else 0) +
                    0.124 * (1 if self._tfidf_score < -2 else 0) +
                    0.120 * (1 if self._tfidf_score < -4 else 0) +
                    0.101 * (1 if self._tfidf_score < -1 else 0) +
                    0.025 * (1 if self._tfidf_score < -8 else 0)
                    )
            # above experiments show that multiple inbound call references are basically always indeed functions
            if len(self.call_ref_sources) > 1:
                self._confidence = 1.0
            # initially recognized candidates are also almost always functions as they follow this heuristic
            elif self.is_initial_candidate:
                self._confidence = round(0.5 + 0.5 * (weighted_confidence), 3)
            else:
                self._confidence = round(weighted_confidence, 3)
        return self._confidence

    def hasCommonFunctionStart(self):
        for length in sorted([int(l) for l in COMMON_PROLOGUES], reverse=True):
            byte_sequence = self.bytes[:length]
            if byte_sequence in COMMON_PROLOGUES["%d" % length][self.bitness]:
                return True
        return False

    def getFunctionStartScore(self):
        if self.function_start_score is None:
            for length in sorted([int(l) for l in COMMON_PROLOGUES], reverse=True):
                byte_sequence = self.bytes[:length]
                if byte_sequence in COMMON_PROLOGUES["%d" % length][self.bitness]:
                    self.function_start_score = COMMON_PROLOGUES["%d" % length][self.bitness][byte_sequence]
                    break
            self.function_start_score = self.function_start_score if self.function_start_score else 0
        return self.function_start_score

    def addCallRef(self, source_addr):
        if source_addr not in self.call_ref_sources:
            self.call_ref_sources.append(source_addr)
        self._score = None

    def removeCallRefs(self, source_addrs):
        for addr in source_addrs:
            if addr in self.call_ref_sources:
                self.call_ref_sources.remove(addr)
        self._score = None

    def setIsTailcallCandidate(self, is_tailcall):
        self.is_tailcall = is_tailcall

    def setInitialCandidate(self, initial):
        self.is_initial_candidate = initial

    def setIsGapCandidate(self, gap):
        self.is_gap_candidate = gap

    def setLanguageSpec(self, lang_spec):
        self.lang_spec = lang_spec
        self._score = None

    def setIsSymbol(self, is_symbol):
        self.is_symbol = is_symbol
        self._score = None

    def setIsExceptionHandler(self, is_exception_handler):
        self.is_exception_handler = is_exception_handler
        self._score = None

    def setIsStub(self, is_stub):
        self.is_stub = is_stub
        self._score = None

    def setAnalysisAborted(self, reason):
        self.finished = True
        self.analysis_aborted = True
        self.abortion_reason = reason

    def setAnalysisCompleted(self):
        self.finished = True

    def isFinished(self):
        return self.finished

    def calculateScore(self):
        score = 0
        score += 10000 if self.is_symbol else 0
        score += 5000 if self.is_exception_handler else 0
        score += 1000 if self.is_stub else 0
        score += 100 if self.lang_spec is not None else 0
        score += self.getFunctionStartScore()
        num_call_refs = len(self.call_ref_sources)
        if num_call_refs >= 10:
            call_ref_score = 10 + int(num_call_refs / 10)
        else:
            call_ref_score = num_call_refs
        score += 10 * call_ref_score
        score += 1 if self.alignment else 0
        return score

    def getScore(self):
        if self._score is None:
            self._score = self.calculateScore()
        return self._score

    def __lt__(self, other):
        own_score = self.getScore()
        other_score = other.getScore()
        if own_score == other_score:
            return self.addr > other.addr
        return own_score < other_score

    def getCharacteristics(self):
        is_aligned = "a" if self.alignment else "-"
        is_finished = "f" if self.finished else "-"
        is_gap = "g" if self.is_gap_candidate else "-"
        is_initial = "i" if self.is_initial_candidate else "-"
        is_lang_spec = "l" if self.lang_spec is not None else "-"
        is_prologue = "p" if self.hasCommonFunctionStart() else "-"
        is_ref = "r" if self.call_ref_sources else "-"
        is_symbol = "s" if self.is_symbol else "-"
        is_tailcall = "t" if self.is_tailcall else "-"
        is_stub = "u" if self.is_stub else "-"
        is_aborted = "x" if self.analysis_aborted else "-"
        characteristics = is_initial + is_symbol + is_stub + is_aligned + is_lang_spec + is_prologue + is_ref + is_tailcall + is_gap + is_finished + is_aborted
        return characteristics

    def __str__(self):
        characteristics = self.getCharacteristics()
        prologue_score = "%d" % self.getFunctionStartScore()
        ref_summary = "{}".format(len(self.call_ref_sources)) if len(self.call_ref_sources) != 1 else "{}: 0x{:x}".format(len(self.call_ref_sources), self.call_ref_sources[0])
        return "0x{:x}: {} -> {} (total score: {}), inref: {} | {}".format(self.addr, hexlify(self.bytes), prologue_score, self.getScore(), ref_summary, characteristics)

    def toJson(self):
        return {
            "addr": self.addr,
            "bytes": self.bytes.hex(),
            "alignment": self.alignment,
            "reason": self.abortion_reason,
            "num_refs": len(self.call_ref_sources),
            "characteristics": self.getCharacteristics(),
            "prologue_score": self.getFunctionStartScore(),
            "score": self.calculateScore(),
            "confidence": self.getConfidence()
        }



import struct
import logging

import lief
lief.logging.disable()

LOG = logging.getLogger(__name__)



class PeFileLoader(object):

    BITNESS_MAP = {0x14c: 32, 0x8664: 64}

    @staticmethod
    def isCompatible(data):
        return data[:2] == b"MZ"

    @staticmethod
    def mapBinary(binary):
        # This is a pretty rough implementation but does the job for now
        mapped_binary = bytearray([])
        pe_offset = PeFileLoader.getPeOffset(binary)
        if pe_offset:
            num_sections = 0
            bitness = 0
            section_infos = []
            optional_header_size = 0xF8
            if pe_offset and len(binary) >= pe_offset + 0x8:
                num_sections = struct.unpack("H", binary[pe_offset + 0x6:pe_offset + 0x8])[0]
                bitness = PeFileLoader.getBitness(binary)
                if bitness == 64:
                    optional_header_size = 0x108
            if pe_offset and num_sections and len(binary) >= pe_offset + optional_header_size + num_sections * 0x28:
                for section_index in range(num_sections):
                    section_offset = section_index * 0x28
                    slice_start = pe_offset + optional_header_size + section_offset + 0x8
                    slice_end = pe_offset + optional_header_size + section_offset + 0x8 + 0x10
                    virt_size, virt_offset, raw_size, raw_offset = struct.unpack("IIII", binary[slice_start:slice_end])
                    section_info = {
                        "section_index": section_index,
                        "virt_size": virt_size,
                        "virt_offset": virt_offset,
                        "raw_size": raw_size,
                        "raw_offset": raw_offset,
                    }
                    section_infos.append(section_info)
            max_virt_section_offset = 0
            min_raw_section_offset = 0xFFFFFFFF
            if section_infos:
                for section_info in section_infos:
                    max_virt_section_offset = max(max_virt_section_offset, section_info["virt_size"] + section_info["virt_offset"])
                    max_virt_section_offset = max(max_virt_section_offset, section_info["raw_size"] + section_info["virt_offset"])
                    if section_info["raw_offset"] > 0x200:
                        min_raw_section_offset = min(min_raw_section_offset, section_info["raw_offset"])
            # support up to 100MB for now.
            if max_virt_section_offset and max_virt_section_offset < 100 * 1024 * 1024:
                mapped_binary = bytearray([0] * max_virt_section_offset)
                mapped_binary[0:min_raw_section_offset] = binary[0:min_raw_section_offset]
            for section_info in section_infos:
                mapped_from = section_info["virt_offset"]
                mapped_to = section_info["virt_offset"] + section_info["raw_size"]
                mapped_binary[mapped_from:mapped_to] = binary[section_info["raw_offset"]:section_info["raw_offset"] + section_info["raw_size"]]
                LOG.debug("Mapping %d: raw 0x%x (0x%x bytes) -> virtual 0x%x (0x%x bytes)",
                          section_info["section_index"],
                          section_info["raw_offset"],
                          section_info["raw_size"],
                          section_info["virt_offset"],
                          section_info["virt_size"])
            LOG.debug("Mapped binary of size %d bytes (%d sections) to memory view of size %d bytes", len(binary), num_sections, len(mapped_binary))
        return bytes(mapped_binary)

    @staticmethod
    def getBitness(binary):
        bitness_id = 0
        pe_offset = PeFileLoader.getPeOffset(binary)
        if pe_offset:
            if pe_offset and len(binary) >= pe_offset + 0x6:
                bitness_id = struct.unpack("H", binary[pe_offset + 0x4:pe_offset + 0x6])[0]
        return PeFileLoader.BITNESS_MAP.get(bitness_id, 0)

    @staticmethod
    def getBaseAddress(binary):
        base_addr = 0
        pe_offset = PeFileLoader.getPeOffset(binary)
        if pe_offset and len(binary) >= pe_offset + 0x38:
            if PeFileLoader.getBitness(binary) == 32:
                base_addr = struct.unpack("I", binary[pe_offset + 0x34:pe_offset + 0x38])[0]
            elif PeFileLoader.getBitness(binary) == 64:
                base_addr = struct.unpack("Q", binary[pe_offset + 0x30:pe_offset + 0x38])[0]
        if base_addr:
            LOG.debug("Changing base address from 0 to: 0x%x for inference of reference counts (based on PE header)", base_addr)
        return base_addr

    @staticmethod
    def getPeOffset(binary):
        if len(binary) >= 0x40:
            pe_offset = struct.unpack("H", binary[0x3c:0x3c + 2])[0]
            return pe_offset
        return 0

    @staticmethod
    def getOEP(binary):
        oep_rva = 0
        if PeFileLoader.checkPe(binary):
            pe_offset = PeFileLoader.getPeOffset(binary)
            if pe_offset and len(binary) >= pe_offset + 0x2c:
                oep_rva = struct.unpack("I", binary[pe_offset + 0x28:pe_offset + 0x2C])[0]
        return oep_rva

    @staticmethod
    def checkPe(binary):
        pe_offset = PeFileLoader.getPeOffset(binary)
        if pe_offset and len(binary) >= pe_offset + 6:
            bitness = struct.unpack("H", binary[pe_offset + 4:pe_offset + 4 + 2])[0]
            return bitness in PeFileLoader.BITNESS_MAP
        return False

    @staticmethod
    def getCodeAreas(binary):
        pefile = lief.parse(bytearray(binary))
        code_areas = []
        base_address = PeFileLoader.getBaseAddress(binary)
        if pefile and pefile.sections:
            for section in pefile.sections:
                if section.characteristics & lief.PE.SECTION_CHARACTERISTICS.MEM_EXECUTE:
                    section_start = base_address + section.virtual_address
                    section_size = section.virtual_size
                    if section_size % 0x1000 != 0:
                        section_size += 0x1000 - (section_size % 0x1000)
                    section_end = section_start + section_size
                    code_areas.append([section_start, section_end])
        return PeFileLoader.mergeCodeAreas(code_areas)

    @staticmethod
    def mergeCodeAreas(code_areas):
        merged_code_areas = sorted(code_areas)
        result = []
        index = 0
        while index < len(merged_code_areas) - 1:
            this_area = merged_code_areas[index]
            next_area = merged_code_areas[index + 1]
            if this_area[1] != next_area[0]:
                result.append(this_area)
                index += 1
            else:
                merged_code_areas = merged_code_areas[:index] + [[this_area[0], next_area[1]]] + merged_code_areas[index + 2:]
        return merged_code_areas



import logging

from .definitions import END_INS, CALL_INS

LOGGER = logging.getLogger(__name__)


class FunctionAnalysisState(object):

    def __init__(self, start_addr, disassembly):
        self.start_addr = start_addr
        self.disassembly = disassembly
        self.block_queue = [start_addr]
        self.current_block = []
        self.blocks = []
        self.num_blocks_analyzed = 0
        self.instructions = []
        self.instruction_start_bytes = set([])
        self.processed_blocks = set([])
        self.processed_bytes = set([])
        self.jump_targets = set([])
        self.call_register_ins = []
        self.block_start = 0xFFFFFFFF
        self.data_bytes = set([])
        self.data_refs = set([])
        self.code_refs = set([])
        self.code_refs_from = {}
        self.code_refs_to = {}
        self.suspicious_ins_count = 0
        self.is_jmp = False
        self.is_next_instruction_reachable = True
        self.is_block_ending_instruction = False
        self.is_sanely_ending = False
        self.has_collision = False
        # set a flag that this tailcall has already been resolved so it does not have to be reanalyzed several times
        self.is_tailcall_function = False
        self.is_leaf_function = True
        self.is_recursive = False
        self.is_thunk_call = False
        self.label = ""

    def chooseNextBlock(self):
        self.is_block_ending_instruction = False
        self.block_start = self.block_queue.pop()
        self.processed_blocks.update([self.block_start])
        return self.block_start

    def addBlockToQueue(self, block_start):
        if not block_start in self.processed_blocks:
            self.block_queue.append(block_start)

    def endBlock(self):
        if self.current_block:
            self.num_blocks_analyzed += 1
            # self.blocks.append(self.current_block)
        self.current_block = []

    def addInstruction(self, i_address, i_size, i_mnemonic, i_op_str, i_bytes):
        ins = (i_address, i_size, i_mnemonic, i_op_str, i_bytes)
        self.instructions.append(ins)
        self.instruction_start_bytes.add(ins[0])
        self.current_block.append(ins)
        for byte in range(i_size):
            self.processed_bytes.add(i_address + byte)
        if self.is_next_instruction_reachable:
            self.addCodeRef(i_address, i_address + i_size, self.is_jmp)
        self.is_jmp = False

    def addCodeRef(self, addr_from, addr_to, by_jump=False):
        self.code_refs.update([(addr_from, addr_to)])
        refs_from = self.code_refs_from.get(addr_from, set([]))
        refs_from.update([addr_to])
        self.code_refs_from[addr_from] = refs_from
        refs_to = self.code_refs_to.get(addr_to, set([]))
        refs_to.update([addr_from])
        self.code_refs_to[addr_to] = refs_to
        if by_jump:
            self.is_jmp = True
            self.jump_targets.update([addr_to])

    def addDataRef(self, addr_from, addr_to, size=1):
        self.data_refs.update([(addr_from, addr_to)])
        for i in range(size):
            self.data_bytes.update([addr_to + i])

    def backtrackInstructions(self, addr_from, num_instructions):
        backtracked = []
        for instruction in sorted(self.instructions, key=lambda x: x[0]):
            if instruction[0] >= addr_from:
                break
            backtracked.append(instruction)
        return backtracked[-num_instructions:]

    def identifyCallConflicts(self, all_refs):
        conflicts = {}
        non_instruction_start_bytes = self.processed_bytes.difference(self.instruction_start_bytes)
        conflict_addrs = set(all_refs.keys()).intersection(non_instruction_start_bytes)
        for candidate_source_ref in conflict_addrs:
            candidate = all_refs[candidate_source_ref]
            if candidate not in conflicts:
                conflicts[candidate] = []
            conflicts[candidate].append(candidate_source_ref)
        return conflicts

    def _finalizeRegularAnalysis(self):
        fn_min = min([ins[0] for ins in self.instructions])
        fn_max = max([ins[0] + ins[1] for ins in self.instructions])

        self.disassembly.function_symbols[self.start_addr] = self.label
        self.disassembly.function_borders[self.start_addr] = (fn_min, fn_max)
        for ins in self.instructions:
            self.disassembly.instructions[ins[0]] = (ins[2], ins[1])
            for offset in range(ins[1]):
                self.disassembly.code_map[ins[0] + offset] = ins[0]
                self.disassembly.ins2fn[ins[0] + offset] = self.start_addr
        for cref in self.code_refs:
            self.disassembly.addCodeRefs(cref[0], cref[1])
        for dref in self.data_refs:
            self.disassembly.addDataRefs(dref[0], dref[1])
        self.disassembly.data_map.update(self.data_bytes)
        self.disassembly.functions[self.start_addr] = self.getBlocks()
        if self.is_recursive:
            self.disassembly.recursive_functions.add(self.start_addr)
        if self.is_leaf_function:
            self.disassembly.leaf_functions.add(self.start_addr)
        if self.is_thunk_call:
            self.disassembly.thunk_functions.add(self.start_addr)

    def finalizeAnalysis(self, as_gap=False):
        if as_gap:
            LOGGER.debug("0x%08x had sanity state: %s (%d ins, blocks: %d)", self.start_addr, self.is_sanely_ending, len(self.instructions), self.num_blocks_analyzed)
            #for instruction in sorted(self.instructions):
            #    print("0x%08x: %s %s" % (instruction[0], instruction[2], instruction[3]))
        if as_gap and not self.is_sanely_ending:
            if len(self.instructions) == 1 and self.instructions[0][2] == "jmp":
                byte = self.disassembly.getByte(self.instructions[0][0])
                if isinstance(byte, int):
                    byte = chr(byte)
                if byte == "\xeb":
                    return False
                # sane case, stub found that just jumps to a referenced function
            elif self.num_blocks_analyzed == 1 and self.instructions[-1][2] in ["jmp", "call"]:
                # similar case to the one above, mostly stubs with tailcalls to a function or shared tail block.
                pass
            else:
                return False
        # in case we have a successful analysis, forward results to disassembly
        if self.num_blocks_analyzed:
            self._finalizeRegularAnalysis()
        return True

    def revertAnalysis(self):
        """
        Remove the analysis results from the disassembly
        """
        self.disassembly.function_borders.pop(self.start_addr, None)
        for ins in self.instructions:
            self.disassembly.instructions.pop(ins[0], None)
            for byte in range(ins[1]):
                self.disassembly.code_map.pop(ins[0] + byte, None)
                self.disassembly.ins2fn.pop(ins[0] + byte, None)
        for cref in self.code_refs:
            self.disassembly.removeCodeRefs(cref[0], cref[1])
        for dref in self.data_refs:
            self.disassembly.removeDataRefs(dref[0], dref[1])
        self.disassembly.functions.pop(self.start_addr, None)

    def getBlocks(self):
        """
        block derivation strategy:
        walk over all potential block starts, which are the start_addr + all "jump" targets (i.e. CFG redirection targets)
        then, for consecutive instructions, break if
        * they have more than 1 outgoing edge
        * the following instruction has more than 1 incoming edge
        """
        if self.blocks:
            return self.blocks
        self.instructions.sort()
        ins = {i[0]:ind for ind, i in enumerate(self.instructions)}
        potential_starts = set([self.start_addr])
        potential_starts.update(list(self.jump_targets))
        blocks = []
        for start in sorted(potential_starts):
            if not start in ins:
                continue
            block = []
            for i in range(ins[start], len(self.instructions)):
                current = self.instructions[i]
                block.append(current)
                # if one code reference is to another address than the next
                if current[0] in self.code_refs_from:
                    if not current[2] in CALL_INS and not i == len(self.instructions) - 1:
                        if any([r != self.instructions[i+1][0] for r in self.code_refs_from[current[0]]]):
                            break
                if not i == len(self.instructions) - 1 and self.instructions[i+1][0] in self.code_refs_to:
                    if len(self.code_refs_to[self.instructions[i+1][0]]) > 1 or self.instructions[i+1][0] in potential_starts:
                        break
                if current[2] in END_INS:
                    break
            if block:
                blocks.append(block)
        self.blocks = blocks
        return self.blocks

    def hasUnprocessedBlocks(self):
        return len(set(self.block_queue).difference(self.processed_blocks)) > 0

    def isProcessed(self, addr):
        return addr in self.processed_bytes

    def isProcessedFunction(self):
        return self.start_addr in self.disassembly.code_map

    def isNextInstructionReachable(self):
        return self.is_next_instruction_reachable

    def setNextInstructionReachable(self, is_reachable):
        self.is_next_instruction_reachable = is_reachable

    def isBlockEndingInstruction(self):
        return self.is_block_ending_instruction

    def isFirstInstruction(self):
        return len(self.instructions) == 0

    def setBlockEndingInstruction(self, is_ending):
        self.is_block_ending_instruction = is_ending

    def setSanelyEnding(self, is_sanely_ending):
        self.is_sanely_ending = is_sanely_ending

    def setCollision(self, is_colliding):
        self.has_collision = is_colliding

    def setRecursion(self, is_recursive):
        self.is_recursive = is_recursive

    def setThunkCall(self, is_thunk_call):
        self.is_thunk_call = is_thunk_call

    def setLeaf(self, is_leaf):
        self.is_leaf_function = is_leaf

    def __str__(self):
        result = "0x{:x} | current: 0x{:x} | blocks: {} | queue: {} | processed: {} | crefs: {} | drefs: {} | suspicious: {} | ending: {}".format(
            self.start_addr,
            self.block_start,
            len(self.getBlocks()),
            ",".join(["0x%x" % b for b in sorted(self.block_queue)]),
            ",".join(["0x%x" % b for b in sorted(list(self.processed_blocks))]),
            len(self.code_refs),
            len(self.data_refs),
            self.suspicious_ins_count,
            self.is_sanely_ending
        )
        return result



#!/usr/bin/python
from io import BytesIO
import re
import struct
import logging
import lief
import sys
from collections import OrderedDict

from smda.common.labelprovider.GoLabelProvider import GoSymbolProvider

LOGGER = logging.getLogger(__name__)


class LanguageAnalyzer(object):

    def __init__(self, disassembly):
        self.disassembly = disassembly
        self.go_resolver = GoSymbolProvider(None)
        self.strings = None

    def validPEHeader(self):
        is_pe = True
        is_pe |= self.disassembly.binary_info.binary[0:2] == "\x4D\x5A"
        if len(self.disassembly.binary_info.binary) > 0x40:
            pe_offset = struct.unpack("I", self.disassembly.binary_info.binary[0x3c:0x40])[0]
            is_pe |= (len(self.disassembly.binary_info.binary) > pe_offset and self.disassembly.binary_info.binary[pe_offset:pe_offset + 2] == "\x50\x45")
        else:
            is_pe = False
        return is_pe

    #SOURCE: https://gist.github.com/geudrik/03152ba1a148d9475e81
    def _getPETimestamp(self):
        try:
            pe_offset = struct.unpack("I", self.disassembly.binary_info.binary[0x3c:0x40])[0]
            # Seek to PE header and read second DWORD
            ts_offset = pe_offset + 8
            return struct.unpack("I", self.disassembly.binary_info.binary[ts_offset:ts_offset + 4])[0]
        except:
            return 0

    def getStrings(self):
        if not self.strings:
            self.strings = [match.group("string") for match in re.finditer(b"(?P<string>[ -~]{6,128})", self.disassembly.binary_info.binary)]
        return self.strings

    def getVisualBasicScore(self):
        #check for the typical import of msvbvm60.dll
        vb_score = 0.0
        if "MSVBVM60.DLL" in self.getStrings():
            vb_score = 0.5
        return vb_score

    def getDotNetScore(self):
        dot_net_score = 0.0
        #check for the typical import of mscorelib.dll and mscoree.dll
        if "mscorelib.dll" in self.getStrings() or "mscoree.dll" in self.getStrings():
            dot_net_score = 0.35
        #check for a functioning PE-Header and typical meta-data
        if self.validPEHeader():
            match = re.search(b"text\x00\x00\x00(?P<dword>[\S\s]{4})(?P<start>[\S\s]{4})", self.disassembly.binary_info.binary)
            if match:
                start_addr = struct.unpack("<I", match.group("start"))[0]
                if start_addr == 0x2000:
                    dot_net_score = max(dot_net_score, 0.8)
                if self.disassembly.getRawBytes(start_addr, 4) == "\xDB\x4D\x00\x79":
                    dot_net_score = max(dot_net_score, 0.9)
        return dot_net_score

    def checkDelphi(self):
        return self.getDelphiScore() > 0.5

    def getDelphiScore(self):
        delphi_score = 0.0
        #Check if Delphi-Locales are present in strings
        if "Borland\\locales" in self.getStrings():
            delphi_score = max(delphi_score, 0.5)
        if self._getPETimestamp() == 0x2a425e19:
            delphi_score = 0.5
        #Find "Delphi-Saved" Strings
        delphi_strings = [
            match.group("string")
            #Regex: <DWORD_LEN_STRING><STRING><TERMINATOR>
            for match in re.finditer(b"(?P<length>[\S\s]{4})(?P<string>[ -~]{6,128})\x00", self.disassembly.binary_info.binary)
            if len(match.group("string")) == struct.unpack("<I", match.group("length"))[0]
        ]
        if len(delphi_strings) > 100:
            delphi_score = max(delphi_score, 0.8)
        return delphi_score

    def getGoScore(self):
        go_score = 0.0
        strings = self.getStrings()
        if any(b"Go build ID" in s for s in strings):
            go_score = max(go_score, 0.6)

        return go_score

    def checkGo(self):
        return self.getGoScore() > 0.5

    def getDelphiObjects(self):
        binary = self.disassembly.binary_info.binary
        base_addr = self.disassembly.binary_info.base_addr
        #Find T-String-Constructs
        #(T-Strings with function-addresses before them)
        #(T-Strings start with a capital T, followed by any capital letter)
        t_objects = {}
        for match in re.finditer(b"(?P<length>.)(?P<t_string>T[A-Z][a-zA-Z0-9]{4,128})", binary):
            if not len(match.group("t_string")) - ord(match.group("length")) in [-3, -2, -1, 0, 1, 2, 3]:  # ord(match.group("length")) != len(match.group("t_string")):
                continue
            t_object_name = match.group("t_string")[:ord(match.group("length"))]
            addresses = []
            t_string_pos = base_addr + match.span()[1] - (len(match.group("t_string")) + len(match.group("length")))
            num_addresses = 0
            address_offset = match.start()
            found_string_ref = False
            # filter address arrays with form <reference>?<len_t_string><t_string>
            if t_string_pos == (struct.unpack("<I", binary[address_offset - 5:address_offset - 1])[0] + 1):
                continue

            while num_addresses < 1000 and address_offset > 4:
                num_addresses += 1
                address_offset -= 4
                address = struct.unpack("<I", binary[address_offset:address_offset + 4])[0]
                if address == t_string_pos:
                    found_string_ref = True
                    LOGGER.debug("object end marker found " + str(t_object_name) + " 0x%08x " % (t_string_pos - base_addr) + "0x%08x" % (t_string_pos))
                    t_objects[t_object_name] = []
                    for address in addresses:
                        if self.disassembly.isAddrWithinMemoryImage(address):
                            t_objects[t_object_name].append(address)
                        else:
                            LOGGER.debug("outside range: 0x%08x", address)
                    break
                #if the first address is invalid, break
                elif num_addresses == 1 and not  self.disassembly.isAddrWithinMemoryImage(address):
                    break
                addresses.append(address)
            if not found_string_ref:
                LOGGER.debug("no object end marker found" + str(t_object_name) + "0x%08x" % (t_string_pos - base_addr) + "0x%08x" % (t_string_pos))
        return t_objects

    def getGoObjects(self):
        self.go_resolver.update(self.disassembly.binary_info)
        return self.go_resolver.getFunctionSymbols()

    def identify(self):
        result = {
            #programming language : probability
            "c/asm": 0.1,  # if no other language matches, c/asm will
            "_count_thiscalls": 0,
            "_count_delphi_objects": 0,
        }
        #DELPHI
        result["delphi"] = self.getDelphiScore()
        if self.checkDelphi():
            t_objects = self.getDelphiObjects()
            functions = sum([len(t_objects[t_string]) for t_string in t_objects])
            # result["_delphi_objects"] = t_objects.keys()
            result["_count_delphi_objects"] = len(t_objects)
            if len(t_objects) > 5 and functions > 10:
                result["delphi"] = max(result.get("delphi", 0), 0.9)
        #.NET
        result[".net"] = self.getDotNetScore()
        #VISUALBASIC
        result["visualbasic"] = self.getVisualBasicScore()
        #GO
        result["go"] = self.getGoScore()
        #C++
        #check if there is a high number of the following patterns
        #in relation to the number off all functions (->the size of the sample)
        patterns = []
        # mov ecx, <stack_offset>; call XYZ
        patterns.append(b"\x8B\x4D[\S\s]\xE8[\S\s]{3}(\x00|\xFF)")
        # mov ecx, <reg>; call XYZ
        patterns.append(b"\x8B[\xC8-\xCF]\xE8[\S\s]{3}(\x00|\xFF)")
        result["c++"] = min(1, 6.0 * sum([len(re.findall(pattern, self.disassembly.binary_info.binary)) for pattern in patterns]) / max(1, len(self.disassembly.functions)))
        result["_count_thiscalls"] = sum([len(re.findall(pattern, self.disassembly.binary_info.binary)) for pattern in patterns])

        #guess the programming language and
        #return dict with probabilities for the use of certain programming languages
        guess = None
        for lang in [key for key in result if not key.startswith("_")]:
            if not (guess and result[guess] > result[lang]):
                guess = lang
        result["_guess"] = guess
        return result



import datetime

from capstone import Cs, CS_ARCH_X86, CS_MODE_32, CS_MODE_64

from smda.DisassemblyResult import DisassemblyResult
from .IdaInterface import IdaInterface

class IdaExporter(object):

    def __init__(self, config, bitness=None):
        self.config = config
        self.ida_interface = IdaInterface()
        self.bitness = bitness if bitness else self.ida_interface.getBitness()
        self.capstone = None
        self.disassembly = DisassemblyResult()
        self.disassembly.smda_version = config.VERSION
        self._initCapstone()

    def _initCapstone(self):
        self.capstone = Cs(CS_ARCH_X86, CS_MODE_32)
        if self.bitness == 64:
            self.capstone = Cs(CS_ARCH_X86, CS_MODE_64)

    def _convertIdaInsToSmda(self, offset, instruction_bytes):
        cache = [i for i in self.capstone.disasm_lite(instruction_bytes, offset)]
        if cache:
            i_address, i_size, i_mnemonic, i_op_str = cache[0]
            smda_ins = (i_address, i_size, i_mnemonic, i_op_str, instruction_bytes)
        else:
            # record error and emit placeholder instruction
            bytes_as_hex = "".join(["%02x" % c for c in bytearray(instruction_bytes)])
            print("missing capstone disassembly output at 0x%x (%s)" % (offset, bytes_as_hex))
            self.disassembly.errors[offset] = {
                "type": "capstone disassembly failure",
                "instruction_bytes": bytes_as_hex
            }
            smda_ins = (offset, len(instruction_bytes), "error", "error", bytearray(instruction_bytes))
        return smda_ins

    def analyzeBuffer(self, binary_info, cb_analysis_timeout=None):
        """ instead of performing a full analysis, simply collect all data from IDA and convert it into a report """
        self.disassembly.analysis_start_ts = datetime.datetime.utcnow()
        self.disassembly.binary_info = binary_info
        self.disassembly.binary_info.architecture = self.ida_interface.getArchitecture()
        if not self.disassembly.binary_info.base_addr:
            self.disassembly.binary_info.base_addr = self.ida_interface.getBaseAddr()
        if not self.disassembly.binary_info.binary:
            self.disassembly.binary_info.binary = self.ida_interface.getBinary()
        if not self.disassembly.binary_info.bitness:
            self.disassembly.binary_info.bitness = self.bitness
        self.disassembly.function_symbols = self.ida_interface.getFunctionSymbols()
        api_map = self.ida_interface.getApiMap()
        for function_offset in self.ida_interface.getFunctions():
            if self.ida_interface.isExternalFunction(function_offset):
                continue
            converted_function = []
            for block in self.ida_interface.getBlocks(function_offset):
                converted_block = []
                for instruction_offset in block:
                    instruction_bytes = self.ida_interface.getInstructionBytes(instruction_offset)
                    smda_instruction = self._convertIdaInsToSmda(instruction_offset, instruction_bytes)
                    converted_block.append(smda_instruction)
                    self.disassembly.instructions[smda_instruction[0]] = (smda_instruction[2], smda_instruction[1])
                    in_refs = self.ida_interface.getCodeInRefs(smda_instruction[0])
                    for in_ref in in_refs:
                        self.disassembly.addCodeRefs(in_ref[0], in_ref[1])
                    out_refs = self.ida_interface.getCodeOutRefs(smda_instruction[0])
                    for out_ref in out_refs:
                        self.disassembly.addCodeRefs(out_ref[0], out_ref[1])
                        if out_ref[1] in api_map:
                            self.disassembly.addr_to_api[instruction_offset] = api_map[out_ref[1]]
                converted_function.append(converted_block)
            self.disassembly.functions[function_offset] = converted_function
            if self.disassembly.isRecursiveFunction(function_offset):
                self.disassembly.recursive_functions.add(function_offset)
            if self.disassembly.isLeafFunction(function_offset):
                self.disassembly.leaf_functions.add(function_offset)
        self.disassembly.analysis_end_ts = datetime.datetime.utcnow()
        return self.disassembly



#!/usr/bin/env python3
import logging
import re
import struct
import codecs

def occurrences(string, sub):
    # https://stackoverflow.com/a/2970542
    count = start = 0
    while True:
        start = string.find(sub, start) + 1
        if start > 0:
            count += 1
        else:
            return count

LOGGER = logging.getLogger(__name__)


class IntelInstructionEscaper:
    """ Escaper to abstract information from disassembled instructions. Based on capstone disassembly. """

    _aritlog_group = [
        "aaa", "aad", "aam", "aas", "adc", "adcx", "add", "adox", "and", "cdq",
        "cdqe", "daa", "das", "dec", "div", "idiv", "imul", "inc", "lzcnt", "mul",
        "mulx", "neg", "not", "or", "popcnt", "rcl", "rcr", "rol", "ror", "sal",
        "salc", "sar", "sbb", "shl", "shld", "shr", "shrd", "sub", "tzcnt", "xadd",
        "xor", "shlx", "shrx", "sarx"
    ]
    _cfg_group = [
        "arpl", "bound", "call", "clc", "cld", "cli", "cmc", "cmova", "cmovae", "cmovb",
        "cmovbe", "cmove", "cmovge", "cmovl", "cmovle", "cmovne", "cmovs", "cmp", "cmps", "cmpsb",
        "cmpsd", "cmpsw", "iret", "iretd", "ja", "jae", "jb", "jbe", "jcxz", "je",
        "jecxz", "jg", "jge", "jl", "jle", "jmp", "jne", "jno", "jnp", "jns",
        "jo", "jp", "jrcxz", "js", "lcall", "ljmp", "loop", "loope", "loopne", "ret",
        "retf", "retfq", "retn", "seta", "setae", "setb", "setbe", "sete", "setg", "setge",
        "setl", "setle", "setne", "setno", "setnp", "setns", "seto", "setp", "sets", "stc",
        "std", "sti", "test"
    ]
    _mem_group = [
        "bsf", "bsr", "bswap", "bt", "btc", "btr", "bts", "cbw", "cmovg", "cmovno",
        "cmovnp", "cmovns", "cmovo", "cmovp", "cmpxchg", "cmpxchg8b", "cqo", "cwd", "cwde", "lahf",
        "lar", "lds", "lea", "les", "lfs", "lgs", "lods", "lodsb", "lodsd", "lodsq",
        "lodsw", "lsl", "lss", "maskmovq", "mov", "movabs", "movnti", "movntq", "movs", "movsb",
        "movsd", "movss", "movsw", "movsx", "movsxd", "movzx", "rdmsr", "sahf", "scas", "scasb",
        "scasd", "scasq", "scasw", "stos", "stosb", "stosd", "stosq", "stosw", "wrmsr", "xabort",
        "xbegin", "xchg", "xlat", "xlatb", "movbe", "clflush"
    ]
    _stack_group = [
        "enter", "leave", "pop", "popad", "popal", "popf", "popfd", "popfq", "push", "pushad",
        "pushal", "pushf", "pushfd", "pushfq"
    ]
    _privileged_group = [
        "clts", "cpuid", "getsec", "hlt", "in", "ins", "insb", "insd", "insw", "int",
        "int1", "int3", "into", "invd", "invlpg", "lfence", "lgdt", "lidt", "lldt", "lmsw",
        "ltr", "mfence", "out", "outs", "outsb", "outsd", "outsw", "pause", "prefetchnta", "prefetcht0",
        "prefetcht1", "prefetcht2", "prefetchw", "rdpmc", "rdtsc", "rsm", "sfence", "sgdt", "sidt", "sldt",
        "smsw", "str", "syscall", "sysenter", "sysexit", "sysret", "ud", "ud2", "ud2b", "wait",
        "wbinvd", "xsaveopt", "ud0",
    ]
    _aes_group = [
        "aesdec", "aesdeclast", "aesenc", "aesenclast", "aesimc", "aeskeygenassist", "crc32", "rdrand", "rdseed", "sha1msg1",
        "sha1msg2", "sha1nexte", "sha1rnds4", "sha256msg1", "sha256msg2", "sha256rnds2", "vaesenc", "vaesenclast", "xcryptcbc", "xcryptcfb",
        "xcryptebc", "xcryptecb", "xcryptofb", "xstorerng", "xsha1", "xsha256", "xcryptctr", "vaesdec", "vaesdeclast"
    ]
    _float_group = [
        "f2xm1", "fabs", "fadd", "faddp", "fbld", "fbstp", "fchs", "fcmovb", "fcmovbe", "fcmove",
        "fcmovnb", "fcmovnbe", "fcmovne", "fcmovnu", "fcmovu", "fcom", "fcomi", "fcomp", "fcompi", "fcompp",
        "fcos", "fdecstp", "fdiv", "fdivp", "fdivr", "fdivrp", "ffree", "fiadd", "ficom", "ficomp",
        "fidiv", "fidivr", "fild", "fimul", "fincstp", "fist", "fistp", "fisttp", "fisub", "fisubr",
        "fld", "fld1", "fldcw", "fldenv", "fldl2e", "fldl2t", "fldlg2", "fldln2", "fldpi", "fldz",
        "fmul", "fmulp", "fnclex", "fninit", "fnop", "fnsave", "fnstcw", "fnstenv", "fnstsw", "fpatan",
        "fprem", "fprem1", "fptan", "frndint", "frstor", "fscale", "fsetpm", "fsin", "fsincos", "fsqrt",
        "fst", "fstp", "fstpnce", "fsub", "fsubp", "fsubr", "fsubrp", "ftst", "fucom", "fucomi",
        "fucomp", "fucompi", "fucompp", "fxam", "fxch", "fxrstor", "fxsave", "fxtract", "fyl2x", "fyl2xp1",
        "fcomip", "fdisi8087_nop", "feni8087_nop", "ffreep", "fucomip",
    ]
    _xmm_group = [
        "addpd", "addps", "addsd", "addss", "addsubpd", "andn", "andnpd", "andnps", "andpd", "andps",
        "cmpeqps", "cmpeqsd", "cmplesd", "cmpltpd", "cmpltps", "cmpltsd", "cmpneqpd", "cmpneqsd", "cmpneqss", "cmpnlepd",
        "cmpnlesd", "cmpnltsd", "cmpps", "cmpsq", "comisd", "comiss", "cvtdq2pd", "cvtdq2ps", "cvtpd2dq", "cvtpd2ps",
        "cvtpi2ps", "cvtps2dq", "cvtps2pd", "cvtps2pi", "cvtsd2si", "cvtsd2ss", "cvtsi2sd", "cvtsi2ss", "cvtss2sd", "cvtss2si",
        "cvttpd2dq", "cvttps2dq", "cvttps2pi", "cvttsd2si", "cvttss2si", "divpd", "divps", "divsd", "divss", "emms",
        "femms", "haddpd", "lddqu", "ldmxcsr", "maxpd", "maxps", "maxsd", "maxss", "minpd", "minps",
        "minsd", "minss", "movapd", "movaps", "movd", "movddup", "movdq2q", "movdqa", "movdqu", "movhlps",
        "movhpd", "movhps", "movlhps", "movlpd", "movlps", "movmskpd", "movmskps", "movntdq", "movntps", "movq",
        "movsldup", "movsq", "movupd", "movups", "mulpd", "mulps", "mulsd", "mulss", "orpd", "orps",
        "pabsd", "pabsw", "packssdw", "packsswb", "packuswb", "paddb", "paddd", "paddq", "paddsb", "paddsw",
        "paddusb", "paddusw", "paddw", "palignr", "pand", "pandn", "pavgb", "pavgw", "pblendw", "pclmulqdq",
        "pcmpeqb", "pcmpeqd", "pcmpeqq", "pcmpeqw", "pcmpestri", "pcmpgtb", "pcmpgtd", "pcmpgtw", "pcmpistri", "pextrb",
        "pextrd", "pextrw", "phaddd", "phsubsw", "pinsrb", "pinsrd", "pinsrq", "pinsrw", "pmaddubsw", "pmaddwd",
        "pmaxsw", "pmaxub", "pminsw", "pminub", "pmovmskb", "pmovsxdq", "pmovzxwd", "pmulhuw", "pmulhw", "pmulld", "pmullw",
        "pmuludq", "popaw", "por", "psadbw", "pshufb", "pshufd", "pshufhw", "pshuflw", "pshufw", "psignw",
        "pslld", "pslldq", "psllq", "psllw", "psrad", "psraw", "psrld", "psrldq", "psrlq", "psrlw",
        "psubb", "psubd", "psubq", "psubsb", "psubsw", "psubusb", "psubusw", "psubw", "punpckhbw", "punpckhdq",
        "punpckhqdq", "punpckhwd", "punpcklbw", "punpckldq", "punpcklqdq", "punpcklwd", "pushaw", "pxor", "rcpps", "rcpss",
        "rorx", "roundsd", "rsqrtps", "shufpd", "shufps", "sqrtpd", "sqrtps", "sqrtsd", "sqrtss", "stmxcsr", "subpd",
        "subps", "subsd", "subss", "ucomisd", "ucomiss", "unpckhpd", "unpckhps", "unpcklpd", "unpcklps", "vaddpd",
        "vaddsd", "vaddss", "vandnpd", "vandpd", "vandps", "vbroadcasti128", "vcmplesd", "vcmpnltsd", "vcmppd", "vcomisd",
        "vcvtdq2pd", "vcvtpd2dq", "vcvtps2pd", "vcvtsd2ss", "vcvtsi2ss", "vcvtss2si", "vcvttpd2dq", "vcvttsd2si", "vdivsd", "verr",
        "verw", "vfmadd132pd", "vfmadd132sd", "vfmadd213sd", "vfmadd213ss", "vfmadd231sd", "vfmadd231ss", "vfmsub132sd", "vfmsub213sd", "vfmsub213ss",
        "vfmsubaddpd", "vfmsubpd", "vfnmadd132sd", "vfnmadd213sd", "vfnmadd231sd", "vhaddpd", "vinserti128", "vldmxcsr", "vmaxsd", "vmaxss",
        "vminpd", "vminsd", "vmovapd", "vmovd", "vmovddup", "vmovdqa", "vmovdqu", "vmovlhps", "vmovntdq", "vmovq",
        "vmovsd", "vmovss", "vmovups", "vmulpd", "vmulps", "vmulsd", "vorpd", "vorps", "vpackssdw", "vpackuswb",
        "vpaddb", "vpaddd", "vpaddq", "vpaddsb", "vpaddsw", "vpaddusb", "vpaddusw", "vpaddw", "vpalignr", "vpand",
        "vpandn", "vpavgb", "vpavgw", "vpblendd", "vpblendw", "vpbroadcastb", "vpclmulqdq", "vpcmpeqb", "vpcmpeqw", "vpcmpgtd",
        "vperm2f128", "vperm2i128", "vpermd", "vpinsrd", "vpmaddwd", "vpmaxsw", "vpmaxub", "vpmovmskb", "vpmullw", "vpor",
        "vprotd", "vprotq", "vpsadbw", "vpshufb", "vpshufd", "vpslld", "vpslldq", "vpsllq", "vpsrad", "vpsraw",
        "vpsrld", "vpsrldq", "vpsrlq", "vpsubd", "vpsubq", "vpsubusb", "vpsubw", "vptest", "vpunpckhbw", "vpunpckhdq",
        "vpunpckhqdq", "vpunpckhwd", "vpunpckldq", "vpxor", "vrcpss", "vroundsd", "vshufps", "vsqrtsd", "vstmxcsr", "vsubps",
        "vsubsd", "vucomisd", "vucomiss", "vunpcklpd", "vxorpd", "vxorps", "vzeroall", "vzeroupper", "xgetbv", "xorpd",
        "xorps",
        "vsubss", "vpmuldq", "vaddsubps", "vcvttsd2usi", "vcvttss2usi", "vmaxps", "vmovaps", "pfcmpge", "kmovb", "mpsadbw",
        "vextracti128", "vpbroadcastd", "vpbroadcastq", "vpcmpeqd", "vpcmpeqq", "vpermq", "vpextrq", "vpinsrq", "vpmuludq", "vpunpcklqdq"
    ]
    _vmx_group = [
        'invrpt', 'invvpid', 'vmcall', 'vmclear', 'vmfunc', 'vmlaunch', 'vmptrld', 'vmptrst', 'vmread', 'vmresume', 'vmwrite', 'vmxoff', 'vmxon'
    ]
    _registers = [
        "al", "bl", "cl", "dl",
        "ah", "bh", "ch", "dh",
        "ax", "bx", "cx", "dx", "sp", "bp", "si", "di",
        "eax", "ebx", "ecx", "edx", "esp", "ebp", "esi", "edi",
        # 64bit
        "rax", "rbx", "rcx", "rdx", "rsp", "rbp", "rsi", "rdi", "rip",
        "r8", "r9", "r10", "r11", "r12", "r13", "r14", "r15",
        "r8d", "r9d", "r10d", "r11d", "r12d", "r13d", "r14d", "r15d",
        "r8w", "r9w", "r10w", "r11w", "r12w", "r13w", "r14w", "r15w",
        "r8b", "r9b", "r10b", "r11b", "r12b", "r13b", "r14b", "r15b",
        "sil", "dil", "bpl", "spl"
    ]
    _segment_registers = [
        # Segment Registers
        "cs", "ds", "es", "fs", "gs", "ss"
    ]
    _extended_registers = [
        # Extended Registers
        "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7", "xmm8", "xmm9", "xmm10", "xmm11", "xmm12", "xmm13", "xmm14", "xmm15",
        "ymm0", "ymm1", "ymm2", "ymm3", "ymm4", "ymm5", "ymm6", "ymm7", "ymm8", "ymm9", "ymm10", "ymm11", "ymm12", "ymm13", "ymm14", "ymm15"
    ]
    _control_registers = [
        # Debug Registers
        "dr0", "dr1", "dr2", "dr3", "dr4", "dr5", "dr6", "dr7",
        # Control Registers
        "cr0", "cr1", "cr2", "cr3", "cr4", "cr5", "cr6", "cr7", "cr8",
        # Test Registers
        "tr3", "tr4", "tr5", "tr6", "tr7"
    ]

    @staticmethod
    def escapeMnemonic(mnemonic):
        mnemonic = mnemonic.split(" ")[-1]
        if mnemonic in IntelInstructionEscaper._aritlog_group:
            return "A"
        elif mnemonic in IntelInstructionEscaper._cfg_group:
            return "C"
        elif mnemonic in IntelInstructionEscaper._mem_group:
            return "M"
        elif mnemonic in IntelInstructionEscaper._stack_group:
            return "S"
        elif mnemonic in IntelInstructionEscaper._privileged_group:
            return "P"
        elif mnemonic in IntelInstructionEscaper._aes_group:
            return "Y"
        elif mnemonic in IntelInstructionEscaper._float_group:
            return "F"
        elif mnemonic in IntelInstructionEscaper._xmm_group:
            return "X"
        elif mnemonic in IntelInstructionEscaper._vmx_group:
            return "V"
        elif mnemonic == "nop":
            return "N"
        elif mnemonic == "error":
            return "U"
        else:
            LOGGER.error("********************************************** Unhandled mnemonic: %s", mnemonic)
            return "U"
        return mnemonic

    @staticmethod
    def escapeField(op_field, escape_registers=True, escape_pointers=True, escape_constants=True):
        op_field = op_field.strip()
        escaped_field = ""
        if op_field == "":
            return ""
        if escape_registers:
            if op_field in IntelInstructionEscaper._registers:
                escaped_field = "REG"
            elif op_field in IntelInstructionEscaper._segment_registers:
                escaped_field = "SREG"
            elif op_field in IntelInstructionEscaper._extended_registers:
                escaped_field = "XREG"
            elif re.search("zmm[0-9]+", op_field):
                escaped_field = "XREG"
            elif op_field in IntelInstructionEscaper._control_registers:
                escaped_field = "CREG"
            elif op_field.startswith("st"):
                escaped_field = "FREG"
            elif op_field.startswith("mm"):
                escaped_field = "MMREG"
        if escape_pointers:
            if (op_field.startswith("xmmword ptr")
                    or op_field.startswith("ymmword ptr")
                    or op_field.startswith("zmmword ptr")
                    or op_field.startswith("xword ptr")
                    or op_field.startswith("tbyte ptr")
                    or op_field.startswith("qword ptr")
                    or op_field.startswith("dword ptr")
                    or op_field.startswith("word ptr")
                    or op_field.startswith("byte ptr")
                    or op_field.startswith("ptr")
                    or op_field.startswith("[")):
                escaped_field = "PTR"
        if escape_constants:
            # potentially include specific constants as extension to CONST
            try:
                op_as_int = int(op_field)
                # if op_as_int in [0, 1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x100, 0xFF, 0xFFFFFFFF, -1]:
                #     escaped_field += "_%d" % op_as_int
                escaped_field = "CONST"
            except:
                pass
            try:
                op_as_int = int(op_field, 16)
                # if op_as_int in [0, 1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x100, 0xFF, 0xFFFFFFFF, -1]:
                #     escaped_field += "_%d" % op_as_int
                escaped_field = "CONST"
            except:
                pass
            if ":" in op_field:
                escaped_field = "CONST"
        if not escaped_field:
            escaped_field = op_field
        return escaped_field

    @staticmethod
    def escapeOperands(ins, offsets_only=False):
        opstring = ins.operands
        op_fields = opstring.split(",")
        esc_regs = True
        esc_consts = True
        if offsets_only:
            if ins.mnemonic in [
                    "call", "lcall", "jmp", "ljmp",
                    "je", "jne", "js", "jns", "jp", "jnp", "jo", "jno", "jl", "jle", "jg",
                    "jge", "jb", "jbe", "ja", "jae", "jcxz", "jecxz", "jrcxz",
                    "loop", "loopne", "loope"]:
                return "OFFSET"
            esc_regs = False
            esc_consts = False
        escaped_fields = []
        for op_field in op_fields:
            escaped_fields.append(IntelInstructionEscaper.escapeField(op_field, escape_registers=esc_regs, escape_constants=esc_consts))
        return ", ".join(escaped_fields)

    @staticmethod
    def escapeBinary(ins, escape_intraprocedural_jumps=False, lower_addr=None, upper_addr=None):
        escaped_sequence = ins.bytes
        # remove segment, operand, address, repeat override prefixes
        if ins.mnemonic in [
                "call", "lcall", "jmp", "ljmp",
                "loop", "loopne", "loope"]:
            escaped_sequence = IntelInstructionEscaper.escapeBinaryJumpCall(ins)
            return escaped_sequence
        if ins.mnemonic in [
                "je", "jne", "js", "jns", "jp", "jnp", "jo", "jno", "jl", "jle", "jg",
                "jge", "jb", "jbe", "ja", "jae", "jcxz", "jecxz", "jrcxz"]:
            escaped_sequence = IntelInstructionEscaper.escapeBinaryJumpCall(ins, escape_intraprocedural_jumps)
            return escaped_sequence
        if "ptr [0x" in ins.operands or "[rip + 0x" in ins.operands or "[rip - 0x" in ins.operands:
            escaped_sequence = IntelInstructionEscaper.escapeBinaryPtrRef(ins)
        if lower_addr is not None and upper_addr is not None and ins.operands.startswith("0x") or ", 0x" in ins.operands:
            immediates = []
            for immediate_match in re.finditer(r"0x[0-9a-fA-F]{1,8}", ins.operands):
                immediate = int(immediate_match.group()[2:], 16)
                if lower_addr > 0x00100000 and lower_addr <= immediate < upper_addr:
                    immediates.append(immediate)
                    escaped_sequence = IntelInstructionEscaper.escapeBinaryValue(escaped_sequence, immediate)
        return escaped_sequence

    @staticmethod
    def escapeBinaryJumpCall(ins, escape_intraprocedural_jumps=False):
        clean_bytes = IntelInstructionEscaper.getByteWithoutPrefixes(ins)
        if escape_intraprocedural_jumps and (
                clean_bytes.startswith("7") or
                clean_bytes.startswith("e0") or
                clean_bytes.startswith("e1") or
                clean_bytes.startswith("e2") or
                clean_bytes.startswith("e3") or
                clean_bytes.startswith("eb")):
            return ins.bytes[:-2] + "??"
        if escape_intraprocedural_jumps and clean_bytes.startswith("0f8"):
            return ins.bytes[:-8] + "????????"
        # these should cover most cross-function references and absolute offsets
        if (clean_bytes.startswith("e8") or
                clean_bytes.startswith("e9")):
            return ins.bytes[:-8] + "????????"
        if clean_bytes.startswith("ff"):
            if len(clean_bytes) <= 8:
                # these seem to be all relative or register based instructions and need no escaping
                return ins.bytes
            if (clean_bytes.startswith("ff14") or
                    clean_bytes.startswith("ff15") or
                    clean_bytes.startswith("ff24") or
                    clean_bytes.startswith("ff25") or
                    clean_bytes.startswith("ffaa")):
                    # FF9*: call dword ptr [<reg> + <offset>] - seem all relative in our test data
                return ins.bytes[:-8] + "????????"
        if clean_bytes.startswith("48"):
            if clean_bytes.startswith("48ff61") and len(clean_bytes) == 8:
                # jmp qword/fword ptr [<register> + <offset>]
                # these are definitely found as interprocedurals but might also be intraprocedurals?
                return ins.bytes[:-2] + "??"
            if clean_bytes.startswith("48ff25"):
                # jmp qword ptr [rip + <offset>]
                return ins.bytes[:-8] + "????????"
        if (clean_bytes.startswith("ea") or
                clean_bytes.startswith("9a")):
                # 9A*: lcall dword ptr [<seg> + <offset>]
                # EA*: ljmp dword ptr [<seg> + <offset>]
            return ins.bytes[:-12] + "????????????"
        return ins.bytes

    @staticmethod
    def escapeBinaryPtrRef(ins):
        escaped_sequence = ins.bytes
        addr_match = re.search(r"\[(rip (\+|\-) )?(?P<dword_offset>0x[a-fA-F0-9]+)\]", ins.operands)
        if addr_match:
            offset = int(addr_match.group("dword_offset"), 16)
            if "rip -" in ins.operands:
                offset = 0x100000000 - offset
            #TODO we need to check if this is actually a 64bit absolute offset (e.g. used by movabs)
            try:
                packed_hex = str(codecs.encode(struct.pack("I", offset), 'hex').decode('ascii'))
            except:
                packed_hex = str(codecs.encode(struct.pack("Q", offset), 'hex').decode('ascii'))
            num_occurrences = occurrences(ins.bytes, packed_hex)
            if num_occurrences == 1:
                escaped_sequence = ins.bytes.replace(packed_hex, "????????")
            elif num_occurrences == 2:
                escaped_sequence = ins.bytes.replace(packed_hex, "????????", 1)
                LOGGER.warning("IntelInstructionEscaper.escapeBinaryPtrRef: 2 occurrences for %s in %s (%s %s), escaping only the first one", packed_hex, ins.bytes, ins.mnemonic, ins.operands)
            elif num_occurrences > 2:
                LOGGER.warning("IntelInstructionEscaper.escapeBinaryPtrRef: more than 2 occurrences for %s", packed_hex)
        return escaped_sequence

    @staticmethod
    def escapeBinaryValue(escaped_sequence, value):
        packed_hex = str(codecs.encode(struct.pack("I", value), 'hex').decode('ascii'))
        num_occurrences = occurrences(escaped_sequence, packed_hex)
        if num_occurrences == 1:
            escaped_sequence = escaped_sequence.replace(packed_hex, "????????")
        elif num_occurrences == 2:
            escaped_sequence = "????????".join(escaped_sequence.rsplit(packed_hex, 1))
            escaped_sequence = "????????".join(escaped_sequence.rsplit(packed_hex, 1))
            LOGGER.warning("IntelInstructionEscaper.escapeBinaryValue: 2 occurrences for %s in %s, escaped both, if they were non-overlapping", packed_hex, escaped_sequence)
        elif num_occurrences > 2:
            LOGGER.warning("IntelInstructionEscaper.escapeBinaryValue: more than 2 occurrences for %s", packed_hex)
        return escaped_sequence

    @staticmethod
    def getByteWithoutPrefixes(ins):
        ins_bytes = ins.bytes
        cleaned = ""
        is_cleaning = True
        for prefix_byte in [ins_bytes[i:i+2] for i in range(0, len(ins_bytes), 2)]:
            if is_cleaning and prefix_byte in ["26", "2e", "36", "3e", "64", "65", "66", "67", "f2", "f3"]:
                continue
            else:
                is_cleaning = False
                cleaned += prefix_byte
        return cleaned



from smda.common.SmdaInstruction import SmdaInstruction


class SmdaBasicBlock:

    smda_function = None
    instructions = None
    offset = None
    length = None

    def __init__(self, instructions, smda_function=None):
        assert isinstance(instructions, list)
        self.smda_function = smda_function
        if instructions:
            self.instructions = instructions
            self.offset = instructions[0].offset
            self.length = len(instructions)

    def getInstructions(self):
        for instruction in self.instructions:
            yield instruction

    @classmethod
    def fromDict(cls, block_dict, smda_function=None):
        smda_block = cls(None)
        smda_block.smda_function = smda_function
        smda_block.instructions = [SmdaInstruction.fromDict(d, smda_function=smda_function) for d in block_dict]
        return smda_block

    def toDict(self):
        return [smda_ins.toDict() for smda_ins in self.instructions]

    def __int__(self):
        return self.offset

    def __str__(self):
        return "0x{:08x}: ({:>4})".format(self.offset, self.length)



import struct
import re
import logging

LOGGER = logging.getLogger(__name__)


class JumpTableAnalyzer(object):
    """ Perform jump table handling.
        There are generally a few typical patterns here:
        A) multiplicative
            cmp     eax, jumptable_size
            ja      loc_default
            mov     eax, ds:off_jumptable[eax*4]
            jmp     eax
        B) additive
            cmp     [ebp+arg_4], jumptable_size
            ja      loc_default
            mov     eax, [ebp+arg_4]
            shl     eax, 2
            add     eax, off_jumptable
            mov     eax, [eax]
            jmp     eax
        C) multiplicative, relative
            cmp     rcx, jumptable_size
            lea     r11, off_jumptable
            movsxd  rcx, ds:(off_jumptable)[r11+rdx*4]
            lea     rcx, [r11+rcx]
            jmp     rcx
    """

    def __init__(self, disassembler):
        self.disassembler = disassembler
        self.disassembly = self.disassembler.disassembly
        self.table_offsets = self._findJumpTables()

    def _findJumpTables(self):
        jumptables = set([])
        for match_offset in re.finditer(b"(\x48|\x4c)\x8d.{5}(.\x63|\x77|.\x89..\x63)", self.disassembly.binary_info.binary):
            rel_table_offset = struct.unpack("I", self.disassembly.getRawBytes(match_offset.start() + 3, 4))[0]
            ins_offset = self.disassembly.binary_info.base_addr + match_offset.start()
            table_offset = ins_offset + rel_table_offset + 7
            if self.disassembly.isAddrWithinMemoryImage(table_offset):
                jumptables.add(table_offset)
        return jumptables

    def _findJumpTableSize(self, backtracked):
        jumptable_size = 0
        for instr in backtracked[::-1]:
            if instr[2].startswith("ret"):
                break
            if instr[2] == "cmp" and re.match(r"[a-z0-9]{2,4}, (([0-9])|(0x[0-9a-f]+))", instr[3]):
                jumptable_size = int(instr[3].split(",")[-1].strip(), base=16) + 1
                # print("  0x%x: found potential jump table size with backtracking: 0x%x (%s %s)" % (instr[0], jumptable_size, instr[2], instr[3]))
                break
        return jumptable_size

    def _directHandler(self, jump_instruction_op_str, state, backtracked):
        register = jump_instruction_op_str.lower()
        data_ref_instruction_addr = None
        off_jumptable = None
        for instr in backtracked[:0:-1]:
            if instr[2] == "mov" and re.match(r"[a-z0-9]{2,3}, dword ptr \[[^ ]+ \+ 0x[0-9a-f]+\]", instr[3]):
                data_ref_instruction_addr = instr[0]
                off_jumptable = self.disassembler.getReferencedAddr(instr[3])
                state.addDataRef(data_ref_instruction_addr, off_jumptable, size=4)
                # print("    0x%x: _directHandler() found potential jump table offset (mov) with backtracking: 0x%x (%s %s)" % (instr[0], off_jumptable, instr[2], instr[3]))
                break
            elif instr[2] == "add" and instr[3].startswith(register):
                data_ref_instruction_addr = instr[0]
                off_jumptable = self.disassembler.getReferencedAddr(instr[3])
                state.addDataRef(data_ref_instruction_addr, off_jumptable, size=4)
                # print("  0x%x: _directHandler() found potential jump table offset (add) with backtracking: 0x%x (%s %s)" % (instr[0], off_jumptable, instr[2], instr[3]))
                break
        return off_jumptable

    def _x64Handler(self, state, backtracked, target_register=None):
        off_jumptable = None
        for instr in backtracked[::-1]:
            if instr[2] == "lea" and re.match(r"[a-z0-9]{2,3}, \[rip (\+|\-) 0x[0-9a-f]+\]", instr[3]):
                if target_register and target_register not in instr[3]:
                    continue
                data_ref_instruction_addr = instr[0]
                offset = self.disassembler.getReferencedAddr(instr[3])
                rip_sign = "+" if "+" in instr[3] else "-"
                if rip_sign == "-":
                    offset = offset * -1
                off_jumptable = instr[0] + instr[1] + offset
                state.addDataRef(data_ref_instruction_addr, off_jumptable, size=4)
                # print("  0x%x: _addHandler() found potential jump table offset (mov) with backtracking: 0x%x (%s %s)" % (instr[0], off_jumptable, instr[2], instr[3]))
                break
        return off_jumptable

    def _getx64BonusOffset(self, backtracked):
        bonus_offset = 0
        for instr in backtracked[::-1][:3]:
            if instr[2] == "mov" and re.match(r"[a-z0-9]{2,3},.*0x[0-9a-f]+\]", instr[3]):
                bonus_offset = self.disassembler.getReferencedAddr(instr[3])
                break
        return bonus_offset

    def _extractDirectTableOffsets(self, jumptable_size, off_jumptable):
        jump_targets = set([])
        if jumptable_size and off_jumptable and self.disassembly.isAddrWithinMemoryImage(off_jumptable):
            for index in range(jumptable_size):
                try:
                    entry = struct.unpack("I", self.disassembly.getBytes(off_jumptable + index * 4, 4))[0]
                    jump_targets.add(entry)
                except:
                    continue
        return sorted(list(jump_targets))

    def _extractRelativeTableOffsets(self, jumptable_size, off_jumptable, alternative_base=None, bonus_offset=0):
        jumptable_size = jumptable_size if jumptable_size else 0xFF
        jump_targets = set([])
        jump_base = alternative_base if alternative_base else off_jumptable
        if jumptable_size and off_jumptable and self.disassembly.isAddrWithinMemoryImage(off_jumptable):
            for index in range(jumptable_size):
                rebased = off_jumptable + bonus_offset - self.disassembly.binary_info.base_addr
                try:
                    entry = struct.unpack("I", self.disassembly.getRawBytes(rebased + index * 4, 4))[0]
                    # check if we are hitting a known jump table
                    if index and (off_jumptable + index * 4) in self.table_offsets:
                        # print("  Hit limit for jump table: 0x%x" % (off_jumptable + index * 4))
                        break
                    if not self.disassembly.isAddrWithinMemoryImage(jump_base + entry):
                        break
                    if entry:
                        target = (jump_base + entry) & self.disassembler.getBitMask()
                        jump_targets.add(target)
                        # state.addDataRef(off_jumptable, rebased + index * 4, size=4)
                    elif not alternative_base:
                        break
                except:
                    continue
        return sorted(list(jump_targets))

    def _resolveExplicitTable(self, jump_instruction_address, state, jumptable_address, jumptable_size=None):
        jumptable_size = jumptable_size if jumptable_size is not None else 0xFF
        jumptable_addresses = []
        bitness = self.disassembly.binary_info.bitness
        entry_size = 4 if bitness == 32 else 8
        if self.disassembly.isAddrWithinMemoryImage(jumptable_address):
            for i in range(jumptable_size):
                if bitness == 32:
                    table_entry = struct.unpack("I", self.disassembly.getBytes(jumptable_address + i * entry_size, entry_size))[0]
                elif bitness == 64:
                    table_entry = struct.unpack("Q", self.disassembly.getBytes(jumptable_address + i * entry_size, entry_size))[0]
                if not self.disassembly.isAddrWithinMemoryImage(table_entry):
                    break
                state.addDataRef(jump_instruction_address, jumptable_address + i * entry_size, size=entry_size)
                jumptable_addresses.append(table_entry)
        return jumptable_addresses

    def getJumpTargets(self, jump_instruction, state):
        jump_instruction_address, jump_instruction_size, jump_instruction_mnemonic, jump_instruction_op_str = jump_instruction
        table_offsets = []
        off_jumptable = None
        backtracked = state.backtrackInstructions(jump_instruction_address, 50)
        backtracked_sequence = "-".join([ins[2] for ins in backtracked[::-1]][:3])
        jumptable_size = self._findJumpTableSize(backtracked)
        # if False and jump_instruction_address:
        #     print("0x%x %s %s -> %s" % (jump_instruction_address, jump_instruction_mnemonic, jump_instruction_op_str, backtracked_sequence))
        if jump_instruction_op_str.startswith("dword ptr [") or jump_instruction_op_str.startswith("qword ptr ["):
            off_jumptable = self.disassembler.getReferencedAddr(jump_instruction_op_str)
            table_offsets = self._resolveExplicitTable(jump_instruction_address, state, off_jumptable, jumptable_size)
        else:
            # 32bit cases typically load into target register directly
            if backtracked_sequence.startswith("mov"):
                off_jumptable = self._directHandler(jump_instruction_op_str, state, backtracked)
                table_offsets = self._extractDirectTableOffsets(jumptable_size, off_jumptable)
            elif backtracked_sequence.startswith("add-movsxd"):
                jumptable_size = self._findJumpTableSize(backtracked)
                off_jumptable = self._x64Handler(state, backtracked)
                alternative_base = 0
                if "rsi" in backtracked[::-1][0][3]:
                    alternative_base = self._x64Handler(state, backtracked, "rsi")
                table_offsets = self._extractRelativeTableOffsets(jumptable_size, off_jumptable, alternative_base=alternative_base)
            elif backtracked_sequence.startswith("lea"):
                jumptable_size = self._findJumpTableSize(backtracked)
                off_jumptable = self._x64Handler(state, backtracked)
                table_offsets = self._extractRelativeTableOffsets(jumptable_size, off_jumptable)
            elif backtracked_sequence.startswith("add-add") or backtracked_sequence.startswith("add-shr"):
                jumptable_size = self._findJumpTableSize(backtracked)
                off_jumptable = self._x64Handler(state, backtracked)
                table_offsets = self._extractRelativeTableOffsets(jumptable_size, off_jumptable)
            elif backtracked_sequence.startswith("add-mov"):
                jumptable_size = self._findJumpTableSize(backtracked)
                off_jumptable = self._x64Handler(state, backtracked)
                bonus = self._getx64BonusOffset(backtracked)
                table_offsets = self._extractRelativeTableOffsets(jumptable_size, off_jumptable, bonus_offset=bonus)
        # if False and off_jumptable and table_offsets:
        #     print("  Found jump table: 0x%x -> %d" % (off_jumptable, len(table_offsets)))
        #     for offset in sorted(list(set(table_offsets))):
        #         print("    0x%x" % offset)
        return table_offsets
